unit codeexp_codegenp;

{ Hidden Paths of Delphi 3, by Ray Lischner.
  Informant Press, 1997.
  Copyright  1997 Tempest Software, Inc.
}

interface

uses Classes, SysUtils, EditIntf, ToolIntf;





procedure ReplaceSelection(Editor: TIEditorInterface;
  ViewNum: Integer; Text: string);

type
  TEditReaderStream = class(TStream)
  private
    fSize: LongInt;
    fPosition: LongInt;
    fReader: TIEditReader;
    function GetSize: LongInt;
  public
    constructor Create(EditIntf: TIEditorInterface);
    destructor Destroy; override;
    function Read(var Buffer; Count: LongInt): LongInt; override;
    function Seek(Offset: LongInt; Origin: Word): LongInt;
        override;
    function Write(const Buffer; Count: LongInt): LongInt; override;
    property Size: LongInt read GetSize;
  end;
  TEditorStrings = class(TStrings)
  private
    fStrings: TStrings;
  protected
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
    function GetObject(Index: Integer): TObject; override;
    procedure Put(Index: Integer; const Str: string); override;
    procedure PutObject(Index: Integer; Obj: TObject); override;
    function GetPosition(Index: Integer): LongInt; virtual;
    function GetCharPos(Index: Integer): TCharPos; virtual;
    property notStrings: TStrings read fStrings;
  public
    constructor Create(Editor: TIEditorInterface);
    destructor Destroy; override;
    procedure LoadFromEditor(Editor: TIEditorInterface);
    procedure SaveToEditor(Editor: TIEditorInterface);
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const Str: string); override;
    function PosToCharPos(Pos: LongInt): TCharPos;
    function CharPosToPos(CharPos: TCharPos): LongInt;
    property Position[Index: Integer]: LongInt read GetPosition;
    property CharPos[Index: Integer]: TCharPos read GetCharPos;
  end;


implementation

uses ExptIntf, VirtIntf;




{ TEditReaderStream }
{ Construct the stream from an editor interface. }
constructor TEditReaderStream.Create(EditIntf: TIEditorInterface);
begin
  inherited Create;
  fReader := EditIntf.CreateReader;
  fSize := -1; { size is unknown }
end;

{ Destroy the stream and free all the interfaces that the stream
  created. }
destructor TEditReaderStream.Destroy;
begin
  fReader.Free;
  inherited Destroy;
end;{ Read from the file stream or the editor. }
function TEditReaderStream.Read(var Buffer; Count: LongInt): LongInt;
const
  MaxCount = 31*1024;
var
  NRead: Integer;
  NRequest: Integer;
  BufPtr: PChar;
begin
  { The initial release of D3 does not handle calls to GetText
    where Count >= 32K. It returns a result equal to Count
    without actually retrieving any text. To circumvent this
    problem, grab buffers of 31K at a time. }
  Result := 0;
  BufPtr := @Buffer;
  while Count > 0 do
  begin
    if Count > MaxCount then
      NRequest := MaxCount
    else
      NRequest := Count;
    NRead := fReader.GetText(fPosition, BufPtr, NRequest);
    Inc(fPosition, NRead);
    Inc(BufPtr, NRead);
    Inc(Result, NRead);
    Dec(Count, NRead);
    { Partially completed read means end-of-buffer, so remember
      the buffer size. If NRead = 0, the position might be past
      the end of file, so save the size only when NRead > 0. }
    if (fSize < 0) and (NRead > 0) and (NRead < NRequest) then
      fSize := fPosition;
  end;
end;

{ Seek to a new position. }
function TEditReaderStream.Seek(Offset: LongInt; Origin: Word):
  LongInt;
begin
  case Origin of
  soFromBeginning:    fPosition := Offset;
  soFromCurrent:      fPosition := fPosition + Offset;
  soFromEnd:          fPosition := Size + Offset;
  else
    raise Exception.CreateFmt('Invalid seek origin, %d', [Origin]);
  end;
  Result := fPosition;
end;

function TEditReaderStream.Write(const Buffer; Count: LongInt):
  LongInt;
begin
  raise Exception.Create('Attempt to write to readonly stream!');
end;{ If the stream user must seek relative to the end of the
  stream, then you need to know the size of the stream.
  There is no simple way to determine this. Instead, use
  a binary search to find a position where a single byte
  read is valid, and a read of the subsequent byte is invalid.
  Since this is such a pain, cache the size after the first call,
  and return the cached size for subsequent calls. }
function TEditReaderStream.GetSize: LongInt;
var
  Hi, Lo, Mid: LongInt;
  Ch: Char;
begin
  if fSize < 0 then
  begin
    Hi := High(LongInt);
    Lo := 0;
    while Lo <= Hi do
    begin
      Mid := (Hi + Lo) div 2;
      if fReader.GetText(Mid, @Ch, 1) = 1 then
        Lo := Mid+1
      else
        Hi := Mid-1;
    end;
    fSize := Lo;
  end;
  Result := fSize;
end;



{ TEditorStrings }
{ Create an edit reader string list. }
constructor TEditorStrings.Create(Editor: TIEditorInterface);
begin
  inherited Create;
  fStrings := TStringList.Create;
  LoadFromEditor(Editor);
end;

destructor TEditorStrings.Destroy;
begin
  notStrings.Free;
  inherited Destroy;
end;

{ Load a string list from an editor interface. Read the edit
  reader as a stream. As each line is added to the string list,
  remember the position of that line in the stream. }
procedure TEditorStrings.LoadFromEditor(Editor: TIEditorInterface);
var
  ERStream: TEditReaderStream;
  StrStream: TStringStream;
  Str: PChar;
  Pos, I: Integer;
begin
  ERStream := TEditReaderStream.Create(Editor);
  try
    StrStream := TStringStream.Create('');
    try
      { Read the entire buffer into StrStream. }
      StrStream.CopyFrom(ERStream, 0);
      { Copy every line from StrStream to the string list. }
      notStrings.Text := StrStream.DataString;

      { Scan the string to find the buffer position of each line. }
      Str := PChar(StrStream.DataString);
      Pos := 0;
      for I := 0 to Count-1 do
      begin
        notStrings.Objects[I] := TObject(Pos);
        Inc(Pos, Length(notStrings[I]));
        if Str[Pos] = #13 then
          Inc(Pos);
        if Str[Pos] = #10 then
          Inc(Pos);
      end;
    finally
      StrStream.Free;
    end;
  finally
    ERStream.Free;
  end;
end;

{ Save the string list to an editor interface. The string list
  does not keep track of specific changes, so replace the entire
  file with the text of the string list. }
procedure TEditorStrings.SaveToEditor(Editor: TIEditorInterface);
var
  Writer: TIEditWriter;
begin
  Writer := Editor.CreateUndoableWriter;
  try
    Writer.DeleteTo(High(LongInt));
    Writer.Insert(PChar(fStrings.Text));
  finally
    Writer.Free;
  end;
end;

{ Get a string. }
function TEditorStrings.Get(Index: Integer): string;
begin
  Result := notStrings[Index]
end;

{ Get an object, which is really the string position. }
function TEditorStrings.GetObject(Index: Integer): TObject;
begin
  Result := notStrings.Objects[Index]
end;

{ Set a string. }
procedure TEditorStrings.Put(Index: Integer; const Str: string);
begin
  notStrings[Index] := Str
end;

{ Set a string's position. }
procedure TEditorStrings.PutObject(Index: Integer; Obj: TObject);
begin
  Objects[Index] := Obj;
end;

{ Return the number of lines in the list. }
function TEditorStrings.GetCount: Integer;
begin
  Result := notStrings.Count
end;

procedure TEditorStrings.Clear;
begin
  notStrings.Clear;
end;

procedure TEditorStrings.Delete(Index: Integer);
begin
  notStrings.Delete(Index)
end;

procedure TEditorStrings.Insert(Index: Integer; const Str: string);
begin
  notStrings.Insert(Index, Str);
end;

{ For convenience, return a position as an integer. }
function TEditorStrings.GetPosition(Index: Integer): LongInt;
begin
  Result := LongInt(notStrings.Objects[Index]);
end;

{ Return a position as a character position. }
function TEditorStrings.GetCharPos(Index: Integer): TCharPos;
begin
  Result := PosToCharPos(GetPosition(Index));
end;

{ Get the buffer position given a character position.
  The character position position specifies a line of text.
  Retrieve the buffer position for the start of that line,
  and add the character index. If the character index is
  past the end of line, return the position of the line
  ending. }
function TEditorStrings.CharPosToPos(CharPos: TCharPos): LongInt;
var
  Text: string;
begin
  { CharPos.Line is 1-based; Strings list is 0-based. }
  Text := notStrings[CharPos.Line-1];
  if CharPos.CharIndex > Length(Text) then
    Result := Position[CharPos.Line-1] + Length(Text)
  else
    Result := Position[CharPos.Line-1] + CharPos.CharIndex;
end;

{ Convert a buffer position to a character position.
  Search for the line such that Pos is between the start
  and end positions of the line. That specifies the line
  number. The char index is the offset within the line.
  If Pos lies within a line ending, return the character
  index of the end of the line.

  Line indices are 1-based, and string list indices are
  0-based, so add 1 to get the true line number.

  Use binary search to locate the desired line quickly. }
function TEditorStrings.PosToCharPos(Pos: LongInt): TCharPos;
var
  Lo, Mid, Hi: Integer;
begin
  Lo := 0;
  Hi := notStrings.Count-1;
  while Lo <= Hi do
  begin
    Mid := (Lo + Hi) div 2;
    if Position[Mid] <= Pos then
      Lo := Mid+1
    else
      Hi := Mid-1
  end;

  Result.Line := Lo;
  if Pos >= Position[Lo-1]+Length(notStrings[Lo-1]) then
    Result.CharIndex := Length(notStrings[Lo-1])
  else
    Result.CharIndex := Pos - Position[Lo-1];
end;


{ ReplaceSelection }
{ Replace the current selection with Text, and select Text. The
  caller must specify the view number where the replacement text
  is scrolled into view. }
procedure ReplaceSelection(Editor: TIEditorInterface;
  ViewNum: Integer; Text: string);
var
  Start, After: TCharPos;
  Strings: TEditorStrings;
  View: TIEditView;
  CursorPos: TEditPos;
  CharPos: TCharPos;

  { Replace a columnar selection with a single string.
    Delete a portion of each line within the block, and insert
    the replacement text on the first line. The Start position
    is the upper left corner. The After position is the lower
    right corner. }
  procedure ReplaceColumns(Start, After: TEditPos; RplText: string);
  var
    EditPos: TEditPos;
    StartCharPos: TCharPos;
    AfterCharPos: TCharPos;
    Line: Integer;
    TmpText: string;
  begin
    { For each line in the block... }
    for Line := Start.Line to After.Line do
    begin
      { Determine the starting and ending character indices
        for the current line. }
      EditPos.Line := Line;
      EditPos.Col := Start.Col;
      View.ConvertPos(True, EditPos, StartCharPos);
      EditPos.Col := After.Col;
      View.ConvertPos(True, EditPos, AfterCharPos);
      { Delete the selected portion of the line and insert the
        replacement text so it starts at the same position.
        Lines numbers are 1-based, and the string list
        is 0-based, so subtract 1 for the line index. Strings
        indices are 1-based, but character position indices
        are 0-based, or add 1 for the character index. The After
        column is inclusive. }
      TmpText := Strings[Line-1];
      Delete(TmpText, StartCharPos.CharIndex+1,
        AfterCharPos.CharIndex-StartCharPos.CharIndex+1);
      Insert(RplText, TmpText, StartCharPos.CharIndex+1);
      Strings[Line-1] := TmpText;

      { Clear RplText so the replacement occurs only on the first line. }
      RplText := '';
    end;
  end;

  { Replace the text between Start and After, inclusive, with Text. }
  procedure ReplaceInclusive;
  var
    Writer: TIEditWriter;
    StartPos, AfterPos: Integer;
  begin
    StartPos := View.CharPosToPos(Start);
    AfterPos := View.CharPosToPos(After);
    Writer := Editor.CreateUndoableWriter;
    try
      { Copy the initial part of the file, up to the selection. }
      Writer.CopyTo(StartPos);
      { Delete the block if there is one to delete. }
      if AfterPos+1 > StartPos then
        Writer.DeleteTo(AfterPos + 1);
      { Insert the replacement text. }
      Writer.Insert(PChar(Text));
      { Copy the rest of the file. }
      Writer.CopyTo(High(LongInt));
    finally
      Writer.Free;
    end;
  end;


begin
  Start := Editor.BlockStart;
  After := Editor.BlockAfter;

  View := Editor.GetView(ViewNum);
  if View <> nil then
  try
    Strings := TEditorStrings.Create(Editor);
    try
      case Editor.BlockType of
      btColumn:
        begin
          ReplaceColumns(TEditPos(Start), TEditPos(After), Text);
          Strings.SaveToEditor(Editor);
        end;
      btInclusive:
        ReplaceInclusive;
      btNonInclusive:
        begin
          Dec(After.CharIndex);
          ReplaceInclusive;
        end;
      btLine:
        begin
          Start.CharIndex := 0;              { start of line }
          After.CharIndex := High(SmallInt); { entire line }
          ReplaceInclusive;
        end;
      else
        raise Exception.CreateFmt('Invalid block type, %d',
                                  [Ord(Editor.BlockType)]);
      end;
    finally
      Strings.Free;
    end;

  { Set the cursor position to the start of the original
    block, which is the position where the replacement
    text starts. }
    if Editor.BlockType = btColumn then
    begin
      CursorPos := TEditPos(Start);
      View.ConvertPos(True, CursorPos, CharPos);
    end
    else
    begin
      CharPos := Start;
      View.ConvertPos(False, CursorPos, Start);
    end;
    View.CursorPos := CursorPos;

    { Select the newly inserted text. }
    Editor.BlockVisible := False;
    Editor.BlockType := btNonInclusive;
    Editor.BlockStart := CharPos;
    Inc(CharPos.CharIndex, Length(Text));
    Editor.BlockAfter := CharPos;
    Editor.BlockVisible := True;

    { Make sure the new cursor position is visible. }
    if (CursorPos.Line < View.TopPos.Line) or
       (CursorPos.Line >= View.TopPos.Line+View.ViewSize.CY) then
    begin
      if CursorPos.Col < View.ViewSize.CX then
        CursorPos.Col := 1;
      View.TopPos := CursorPos;
    end;
  finally
    View.Free;
  end;
end;

end.